mod readmap_helper;
use crossterm::event::KeyCode;
use readmap::types::MemoryMap;
use std::path::PathBuf;
use tool_helper::Error;
use ratatui::{
    layout::{Alignment, Constraint, Direction, Layout, Rect},
    style::{Color, Modifier, Style},
    text::{Line, Span},
    widgets::{Block, Borders, BorderType, Gauge, List, ListItem, ListState, Paragraph, Tabs}
};

pub type EventReceiver = std::sync::mpsc::Receiver<Event<crossterm::event::KeyEvent>>;
pub type Terminal      = ratatui::Terminal<ratatui::backend::CrosstermBackend<std::io::Stdout>>;
type ConsoleFrame<'a>  = ratatui::Frame<'a, ratatui::backend::CrosstermBackend<std::io::Stdout>>;

pub enum Event<I> {
    Input(I),
    Tick,
    ForceRender,
}

pub enum UIState {
    Alive,
    Render,
    Terminated
}

enum MenuSelection {
    Stats,
    Quit,
}

impl MenuSelection {
    fn next(&self) -> MenuSelection {
        match self {
            Self::Stats => Self::Quit,
            Self::Quit  => Self::Stats
        }
    }

    fn prev(&self) -> MenuSelection {
        self.next()
    }

    fn get_id(&self) -> usize {
        match self {
            MenuSelection::Stats => 0,
            MenuSelection::Quit  => 1,
        }
    }
}

#[derive(Copy, Clone)]
enum SectionMode {
    UpperSection,
    LowerSection,
}

impl SectionMode {
    pub fn as_id(&self) -> usize {
        match self {
            SectionMode::UpperSection => 0,
            SectionMode::LowerSection => 1,
        }
    }
}

struct SectionSelection {
    mode:     SectionMode,
    id:       [usize; 2]
}

impl SectionSelection {
    fn increment(&mut self, max: usize) {
        let cur_id        = &mut self.id[self.mode.as_id()];
        let new_selection = *cur_id + 1;

        *cur_id = {
            if new_selection >= max {
                0
            }

            else {
                new_selection
            }
        } 
    }

    fn decrement(&mut self, max: usize) {
        let cur_id        = &mut self.id[self.mode.as_id()];
        let new_selection = *cur_id - 1;

        *cur_id = {
            if new_selection >= max {
                max - 1
            }

            else {
                 new_selection
            }
        }
    }

    fn switch_section_to(&mut self, section_mode: SectionMode) {
        self.mode = section_mode;
    }

    fn get_section_mode_id(&self) -> usize {
        self.mode.as_id()
    }

    fn get_selection_for(&self, section_mode: SectionMode) -> usize {
        self.id[section_mode.as_id()]
    }

    fn get_current_selection(&self) -> usize {
        self.get_selection_for(self.mode)
    }
}

impl Default for SectionSelection {
    fn default() -> Self {
        Self{mode: SectionMode::UpperSection, id: [0, 0]}
    }
}

pub struct ConsoleUI {
    data:              ConsoleUIData,
    rx:                EventReceiver,
    terminal:          Terminal,
    console_size:      Rect,
    menu_selection:    MenuSelection,
    section_selection: SectionSelection,
}

impl ConsoleUI {
    pub fn new(rx: EventReceiver, terminal: Terminal) -> ConsoleUI {
        ConsoleUI{data: ConsoleUIData::default(), rx, terminal, console_size: Rect::default(), menu_selection: MenuSelection::Stats, section_selection: SectionSelection::default()}
    }

    pub fn update_data(&mut self, memory_map: &MemoryMap) -> Result<(), Error> {
        self.data              = ConsoleUIData::from(memory_map)?;
        self.section_selection = SectionSelection::default();

        Ok(())
    }

    pub fn update(&mut self) -> Result<UIState, Error> {
        let input_result = {
            match self.rx.recv()? {
                Event::Input(event) => {
                    if matches!(self.menu_selection, MenuSelection::Stats) {
                        match event.code {
                            KeyCode::Down => {
                                self.section_selection.increment(self.data.section_info.len());
                                return Ok(UIState::Render);
                            },
                            KeyCode::Up => {
                                self.section_selection.decrement(self.data.section_info.len());
                                return Ok(UIState::Render);
                            },
                            KeyCode::Char('a') => {
                                self.section_selection.switch_section_to(SectionMode::UpperSection);
                                return Ok(UIState::Render);
                            },
                            KeyCode::Char('s') => {
                                self.section_selection.switch_section_to(SectionMode::LowerSection);
                                return Ok(UIState::Render);
                            },
                            _ => ()
                        }
                    }

                    match event.code {
                        KeyCode::Char('q') => Ok(UIState::Terminated),
                        KeyCode::Char('s') => {
                            self.menu_selection = MenuSelection::Stats;
                            Ok(UIState::Render)
                        },
                        KeyCode::Left => {
                            self.menu_selection = self.menu_selection.prev();
                            Ok(UIState::Render)
                        },
                        KeyCode::Right => {
                            self.menu_selection = self.menu_selection.next();
                            Ok(UIState::Render)
                        },
                        KeyCode::Enter => {
                            if matches!(self.menu_selection, MenuSelection::Quit) {
                                Ok(UIState::Terminated)
                            }
                            else {
                                Ok(UIState::Alive)
                            }
                        },
                        _ => Ok(UIState::Alive)
                    }
                },
                Event::Tick => {
                    let terminal_size = self.terminal.size().expect("Getting size of terminal");
                    if terminal_size != self.console_size {
                        self.console_size = terminal_size;
                        Ok(UIState::Render)
                    } 

                    else {
                        Ok(UIState::Alive)
                    }
                },
                Event::ForceRender => Ok(UIState::Render)
            }
        };

        input_result
    }

    pub fn render(&mut self) -> Result<(), Error> {
        self.terminal.draw(|frame| {
            // Create layout elements
            let main_layout = Layout::default().direction(Direction::Vertical).constraints([
                Constraint::Length(3),
                Constraint::Min(3)
            ]).split(frame.size()).to_vec();

            // Elements
            let menu_bar = {
                const MENU_TITLES: &'static [&'static str] = &["Stats", "Quit"];

                let menu = MENU_TITLES.iter().map(|t| {
                    let (first, rest) = t.split_at(1);
                    Line::from(vec![
                        Span::styled(first, Style::default().fg(Color::Yellow).add_modifier(Modifier::UNDERLINED),),
                        Span::styled(rest,  Style::default().fg(Color::White)),
                    ])
                }).collect();
                
                Tabs::new(menu).select(self.menu_selection.get_id()).block(
                    Block::default()
                    .title("psxreadmap - Menu").borders(Borders::ALL))
                    .style(Style::default().fg(Color::White))
                    .highlight_style(Style::default().bg(Color::LightYellow))
                    .divider(Span::raw("|")
                )
            };

            // Render screen
            frame.render_widget(menu_bar, main_layout[0]);
            match self.menu_selection {
                MenuSelection::Stats => {
                    Self::render_stats(frame, main_layout[1], &self.section_selection, &mut self.data);
                },
                MenuSelection::Quit  => {
                    frame.render_widget(Self::create_exit_message(), main_layout[1]);
                }
            }
        })?;
        Ok(())
    }

    pub fn close(mut self) -> Result<(), Error> {
        crossterm::terminal::disable_raw_mode()?;
        self.terminal.show_cursor()?;
        self.terminal.clear()?;

        Ok(())
    }

    fn render_stats(frame: &mut ConsoleFrame, layout: Rect, section_selection: &SectionSelection, data: &mut ConsoleUIData) {
        const HIGHLIGHT_COLOR:[Color;2] = [Color::Cyan, Color::Red];
        let layout = Layout::default().direction(Direction::Vertical).constraints([
            Constraint::Length(3),
            Constraint::Max(3),
            Constraint::Max(3),
        ]).split(layout).to_vec();
        let content_list = List::new({
            let mut vec = Vec::new();

            for section in &data.section_info {
                vec.push(ListItem::new(section.name.clone()))
            }
            vec
        }).style(Style::default().fg(Color::White))
        .block(Block::default()
            .borders(Borders::ALL)
            .style(Style::default().fg(Color::White))
            .border_type(BorderType::Plain)
        )
        .highlight_style(Style::default().bg(HIGHLIGHT_COLOR[section_selection.get_section_mode_id()]));

        let first_mem_gauge  = Self::create_mem_gauge_from(&data.section_info[section_selection.get_selection_for(SectionMode::UpperSection)], "a", HIGHLIGHT_COLOR[0]);
        let second_mem_gauge = Self::create_mem_gauge_from(&data.section_info[section_selection.get_selection_for(SectionMode::LowerSection)], "s", HIGHLIGHT_COLOR[1]);
        let mut list_state = ListState::default();
        
        list_state.select(Some(section_selection.get_current_selection()));

        frame.render_stateful_widget(content_list, layout[0], &mut list_state);
        frame.render_widget(first_mem_gauge, layout[1]);
        frame.render_widget(second_mem_gauge, layout[2]);
    }

    fn create_mem_gauge_from<'a>(section_info: &SectionInfo, string: &'a str, color: Color) -> Gauge<'a> {
        const PERCISION:usize         = 5;
        const FLOAT_DISPLAY_LIMIT:f64 = {
            let mut base = 1.0f64;
            let mut i    = 0;

            while i < PERCISION {
                base = base/10.0f64;
                i += 1;

            }
            base
        };

        let adr_ratio = section_info.get_size_ratio();
        Gauge::default().block(Block::default().borders(Borders::ALL)
            .title(Line::from(vec![
                Span::styled(string, Style::default().fg(color)),
                Span::raw(format!(": {} @0x{:x} - 0x{:x} [{}b/{}b]", 
                section_info.name, section_info.start_adr, section_info.get_end_adr(), section_info.size, section_info.max_size)),
            ])))
            .gauge_style(Style::default().fg(color).bg(Color::Black).add_modifier(Modifier::BOLD))
            .ratio(adr_ratio).label({
                let adr_ratio = adr_ratio*100.0f64;
                if adr_ratio > FLOAT_DISPLAY_LIMIT {
                    format!("{:.width$}%", adr_ratio, width=PERCISION)
                }

                else {
                    format!("< {:.width$}%", FLOAT_DISPLAY_LIMIT, width=PERCISION)
                }
            })
    }

    fn create_exit_message<'a>() -> Paragraph<'a> {
        Paragraph::new("Press \"ENTER\" to exit")
            .style(Style::default().fg(Color::White))
            .alignment(Alignment::Center)
            .block(Block::default()
                .borders(Borders::ALL)
                .style(Style::default().fg(Color::White))
                .border_type(BorderType::Plain).title("QUIT")
            )
    }
}

#[derive(Default)]
struct ConsoleUIData {
    section_info: Vec<SectionInfo>
}

impl ConsoleUIData {
    pub const RAM_BASE_ADDRESS:u64    = 0x80000000;
    pub const RAM_SIZE:usize          = 2*1024*1024;

    pub fn from(memory_map: &MemoryMap) -> Result<Self, Error> {
        fn get_last_section_end_adr(memory_map: &MemoryMap) -> u64 {
            if let Some(section) = memory_map.sections.last() {
                return section.adr + section.size as u64;
            }
            0u64
        }

        let mut new_self     = ConsoleUIData::default();
        let     highest_adr  = get_last_section_end_adr(memory_map);
        let     program_size = (highest_adr - Self::RAM_BASE_ADDRESS) as usize;

        new_self.section_info.push(SectionInfo{
            name:      String::from("<overall usage>"),
            start_adr: Self::RAM_BASE_ADDRESS,
            size:      program_size,
            max_size:  Self::RAM_SIZE
        });

        for section in &memory_map.sections {
            new_self.section_info.push(SectionInfo{
                name:      section.name.clone(),
                start_adr: section.adr,
                size:      section.size,
                max_size:  program_size
            });
        }

        Ok(new_self)
    }
}

#[derive(Default)]
struct SectionInfo {
    pub name:      String,
    pub start_adr: u64,
    pub size:      usize,
    pub max_size:  usize,
}

impl SectionInfo {
    fn get_end_adr(&self) -> u64 {
        self.start_adr + self.size as u64
    }

    fn get_size_ratio(&self) -> f64 {
        self.size as f64 / self.max_size as f64
    }
}

pub fn load_memory_map(use_wsl: bool, input: PathBuf) -> Result<MemoryMap, Error> {
    readmap_helper::generate_memory_map(use_wsl, input)
}